 #
 #  ES1: A C++ Design Metrics Analyzer
 #  Copyright (C) 2000-2001 National Research Council of Canada
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
 #  the Free Software Foundation; either version 2 of the License, or
 #  (at your option) any later version.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU General Public License for more details.
 #
 #  You should have received a copy of the GNU General Public License
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 #



# the next line restarts using wish \
#exec /usr/snavigator/bin/hyper "$0" "$@"

# User defines whether he wants to analyze public, protected or private attributes 
# by setting a flag on the command line. proc open_database opens a .iv file 
# and calls appropriate procedures for extracting private, public or protected 
# attributes. 


#
#  This procedure selects only lines that correspond to public 
#  attributes or methods from the database extracted from the 
#  .iv or .md file
#
#
proc select_public {db_var} {
    set pub_var [$db_var seq -col [list 0 1 2 3 4 "5 :0:4" 6 7 8]]
    return $pub_var
}

#
#  This procedure selects only lines that correspond to protected 
#  attributes or methods from the database extracted from the 
#  .iv or .md file
#
#  
proc select_protected {db_var} {
    set prot_var [$db_var seq -col [list 0 1 2 3 4 "5 :0:2" 6 7 8]]
    return $prot_var
}

#
#  This procedure selects only lines that correspond to private 
#  attributes or methods from the database extracted from the 
#  .iv or .md file
#
#  
proc select_private {db_var} {
    set priv_var [$db_var seq -col [list 0 1 2 3 4 "5 :0:1" 6 7 8]]
    return $priv_var
}



#
#  This procedure is supposed to extract the field "type" of each attribute
#  from the list of attributes (which can be private, public or protected).
#  After that, an effort is made to find the real type : if it is already 
#  a class type, than the procedure check_type_for_classes would note the 
#  coupling between the class that is the attribute's type and the class 
#  that declares that attribute.
#
#  If it is not a class type, that it could be a typedef, and the procedure
#  for determining the type of the typedef is called.
#
#  attribute_list passed as an argument is returned by select_private, 
#  select_public or select_protected procedures. 
#

proc extract_attribute_type {attribute_list} {

    if {$attribute_list == {}} {
	setLog "No attributes in the project."
	return
    }

    puts "Analyzing attribute types ... (please wait)"

    
    foreach entry_v $attribute_list {
#puts -nonewline "."
	set class_c [lindex $entry_v 0]
	set attribute_name [lindex $entry_v 1]
	set attribute_pos [lindex $entry_v 2]
	set attribute_line_no [lindex [extract_line_numbers [list $attribute_pos {}]] 0]

#EXTRACT_START_END_POSITION SHOULD RETURN ALL THE CLASSES WITH THE SAME NAME. AFTER THAT, WE HAVE TO CHECK WHETHER OUR ATTRIBUTE BELONGS TO ONE OR ANOTHER CLASS : we are doing that in foreach element start_end_pos

	set attribute_filename [lindex $entry_v 3]

#we want to know class' start and end position : entry_v gives us only attribute's
	set start_end_pos [extract_start_end_pos $class_c $attribute_filename]

	set attribute_type [lindex $entry_v 6]
	set attribute_type_good [extract_good_attribute_type $attribute_type]

	
#if extract_good_attributes returns an empty string, it means that the 
#type is primitive... so go to the other type

	if {[string compare $attribute_type_good ""] != 0} {
#if there is more than one class with the same name as our attribute's class
	    if {[llength $start_end_pos] > 1} {
#for every such class, we have to check whether our attribute is in it or not		    
		foreach element $start_end_pos {
		    set start_end_line [extract_line_numbers [list $element {}]]
		    set start_line [lindex $start_end_line 0]
		    set end_line [lindex $start_end_line 1]

		    if {$attribute_line_no > $start_line && $attribute_line_no < $end_line} {
			#here we have to check whether our attribute's class is inner. If yes, we have to find its outmost class, we don't consider coupling with inner classes.
			set outer_class_info [is_inner_class $start_line $end_line $attribute_filename]
			if {$outer_class_info != 0} {
			    #set the outer_class, and find the real type
			    if {$attribute_type_good != 0} {
				set_array_of_attributes attributes $class_c $outer_class_info "" $attribute_type_good -1 -1 "ca" "" {} 0
				find_real_type attributes
			    } else {
				#use attribute_type instead of attribute_type_good, which is 0 in case of method pointers
				set_array_of_attributes attributes $class_c $outer_class_info "" $attribute_type -1 -1 "ca" "" {} 0
				analyze_method_pointers attributes
			    }
			} else {
			    #if it's not an inner class, than we use it for coupling measurements
			    set class_c_info [list $class_c [lindex $element 0] [lindex $element 1] $attribute_filename]
			    if {$attribute_type_good != 0} {
				set_array_of_attributes attributes "" $class_c_info "" $attribute_type_good -1 -1 "ca" "" {} 0
				find_real_type attributes
			    } else {
				#use attribute_type instead of attribute_type_good, which is 0 in case of method pointers
				set_array_of_attributes attributes "" $class_c_info "" $attribute_type -1 -1 "ca" "" {} 0
				analyze_method_pointers attributes
			    }
			}
		    }
		}
	    } else { ;#if there is only one class with our attribute's class name

#we have to find out whether it's inner or not, just in case
		set start_end_line [extract_line_numbers $start_end_pos]		
		set start_line [lindex $start_end_line 0]
		set end_line [lindex $start_end_line 1]
		set outer_class_info [is_inner_class $start_line $end_line $attribute_filename]
		if {$outer_class_info != 0} { ;#if it's inner class
		    if {$attribute_type_good != 0} {
			set_array_of_attributes attributes $class_c $outer_class_info "" $attribute_type_good -1 -1 "ca" "" {} 0
			find_real_type attributes
		    } else { ;#if it's a method pointer
			set_array_of_attributes attributes $class_c $outer_class_info "" $attribute_type -1 -1 "ca" "" {} 0
			analyze_method_pointers attributes
		    }			
		} else { ;#if it's not inner class
		    set start_end [lindex $start_end_pos 0]
		    set class_c_info [list $class_c [lindex $start_end 0] [lindex $start_end 1] $attribute_filename]
		    if {$attribute_type_good != 0} {
			set_array_of_attributes attributes "" $class_c_info "" $attribute_type_good -1 -1 "ca" "" {} 0
			find_real_type attributes
		    } else { ;#if it's a method pointer
			set_array_of_attributes attributes "" $class_c_info "" $attribute_type -1 -1 "ca" "" {} 0
			analyze_method_pointers attributes
		    }
		}
	    }
	}
    }
}


#
# This procedure helps setting the arguments for various procedures 
# (find_real_type, check_type ... , analyze_templates ...). It's just
# a helper procedure to lower the complexity of this code. 
#

proc set_array_of_attributes {attributes inner_cl original_cl_info my_class_info attribute_type index_colon index_templates coupling_type_id method_coupling_type_id templ_argument global_argument} {
    
    upvar $attributes array_of_attributes

    set array_of_attributes(inner_class) $inner_cl
    set array_of_attributes(original_class_info) $original_cl_info
    set array_of_attributes(class_c_info) $my_class_info
    set array_of_attributes(attribute_type_good) $attribute_type
    set array_of_attributes(index_of_colon) $index_colon
    set array_of_attributes(index_of_less_than) $index_templates
    set array_of_attributes(coupling_type) $coupling_type_id
    set array_of_attributes(method_coupling_type) $method_coupling_type_id
    set array_of_attributes(t_argument) $templ_argument
    set array_of_attributes(global_arg) $global_argument
}



#
# This procedure checks whether a class is inner, by comparing
# its start and end line numbers with line numbers of all classes
# in that file.
#
# It returns the outer class, with its line numbers, if applicable.
# If not, it returns 0.
#
#

proc is_inner_class {my_start_line my_end_line filename} {
    global db_classe
    set classes_in_my_file [$db_classe seq -col [list 0 1 3 "2 =$filename"]]

    foreach element $classes_in_my_file {
	set start_line [lindex $element 1]
	set start_line [lindex [extract_line_numbers [list $start_line {}]] 0]
	set end_line [lindex $element 2]
	set end_line [lindex [extract_line_numbers [list $end_line {}]] 0]
	if {($start_line < $my_start_line)  && ($end_line > $my_end_line)} {
	    return $element
	}
    }
    return 0
}



#
#  This procedure returns 1 if a typedef (see check_type_for_typedefs)
#  is in a certain class, and 0 if it is not. If it is in an inner class
#  of that class, it returns 0. 
#


proc is_inside_my_class {class_filename start_line end_line line_no} {
    global db_classe

    set min_start_line_for_inner_cl [expr $start_line + 1]
    set max_end_line_for_inner_cl [expr $end_line - 1]

    set my_inner_classes [$db_classe seq -col [list 0 1 "3 -$min_start_line_for_inner_cl:$max_end_line_for_inner_cl" "2 =$class_filename"]]

    foreach inner_class $my_inner_classes {
	set inner_start_line [lindex $inner_class 1]
	set inner_start_line [lindex [extract_line_numbers [list $inner_start_line {}]] 0]
	set inner_end_line [lindex $inner_class 2]
	set inner_end_line [lindex [extract_line_numbers [list $inner_end_line {}]] 0]
	if {($line_no > $inner_start_line) && ($line_no < $inner_end_line)} {
	    return 0
	}
    }
    if {($line_no > $start_line) && ($line_no < $end_line)} {
	return 1
    }
    return 0
}


#
# This procedure extracts method return and argument types from the list returned by 
# select_public/private/protected procedure. For the return type, it tries to find its 
# real type by calling find_real_type. For method arguments the procedure 
# process_list_of_arguments is called. As for attributes, a check for inner classes is performed.
#


proc extract_method_return_and_argument_types {method_list} {

    if {$method_list == {}} {
	setLog "No methods in the project."
	return
    }

puts "Analyzing method return and types ... (please wait)"

    foreach entry_m $method_list {
#puts -nonewline "."
	set class_c [lindex $entry_m 0]
	set list_of_arguments [lindex $entry_m 7]
	set method_name [lindex $entry_m 1]
	set method_pos [lindex $entry_m 2]
	set method_line_no [lindex [extract_line_numbers [list $method_pos {}]] 0]
	set method_filename [lindex $entry_m 3]

	set start_end_pos [extract_start_end_pos $class_c $method_filename]
	
	set return_type [lindex $entry_m 6]
	set return_type_good [extract_good_attribute_type $return_type]

#if extract_good_attribute_type returns an empty string, it means that the 
#type is primitive .... so go to the other type

	if {[string compare $return_type_good ""] != 0} {
	    if {[llength $start_end_pos] > 1} {

#for each class with our class'  name in this file
		foreach element $start_end_pos {
		    set start_end_line [extract_line_numbers [list $element {}]]
		    set start_line [lindex $start_end_line 0]
		    set end_line [lindex $start_end_line 1]
#if method is inside the class
		    if {$method_line_no > $start_line && $method_line_no < $end_line} {
#check if our class is inner class or not
			set outer_class_info [is_inner_class $start_line $end_line $method_filename]
#if our class is inner class, set the outer class
			if {$outer_class_info != 0} {
			    if {$return_type_good != 0} {
				set_array_of_attributes attributes $class_c $outer_class_info "" $return_type_good -1 -1 "cm" "r" {} 0
				find_real_type attributes
			    } else {
#use return_type instead of return_type_good for method pointers
				set_array_of_attributes attributes $class_c $outer_class_info "" $return_type -1 -1 "cm" "r" {} 0
				analyze_method_pointers attributes
			    }
			    set_array_of_attributes attributes $class_c $outer_class_info "" $list_of_arguments -1 -1 "cm" "a" {} 0
			    process_list_of_arguments attributes
			} else {
			    #if our class is not inner class
			    set class_c_info [list $class_c [lindex $element 0] [lindex $element 1] $method_filename]
			    if {$return_type_good != 0} { ;# find the real type
				set_array_of_attributes attributes "" $class_c_info "" $return_type_good -1 -1 "cm" "r" {} 0
				find_real_type attributes
			    } else { ;# it's a method pointer
				set_array_of_attributes attributes "" $class_c_info "" $return_type -1 -1 "cm" "r" {} 0
				analyze_method_pointers attributes
			    }
			    set_array_of_attributes attributes "" $class_c_info "" $list_of_arguments -1 -1 "cm" "a" {} 0
#call to a procedure for analyzing method arguments
			    process_list_of_arguments attributes
			}
		    }
		}
	    } else { ;# if there is only one class with this name
#we have to check whether it's inner or not, just in case.
		set start_end_line [extract_line_numbers $start_end_pos]
		set start_line [lindex $start_end_line 0]
		set end_line [lindex $start_end_line 1]
		set outer_class_info [is_inner_class $start_line $end_line $method_filename]
		if {$outer_class_info != 0} { ;#if it's inner, set the outer class
		    if {$return_type_good != 0} {
			set_array_of_attributes attributes $class_c $outer_class_info "" $return_type_good -1 -1 "cm" "r" {} 0
			find_real_type attributes
		    } else { ;#if it's a method pointer
			set_array_of_attributes attributes $class_c $outer_class_info "" $return_type -1 -1 "cm" "r" {} 0
			analyze_method_pointers attributes
		    }
		    set_array_of_attributes attributes $class_c $outer_class_info "" $list_of_arguments -1 -1 "cm" "a" {} 0
		    process_list_of_arguments attributes
		} else { ;# if it's not inner
		    set start_end [lindex $start_end_pos 0]
		    set class_c_info [list $class_c [lindex $start_end 0] [lindex $start_end 1] $method_filename]
		    if {$return_type_good != 0} {
			set_array_of_attributes attributes "" $class_c_info "" $return_type_good -1 -1 "cm" "r" {} 0
			find_real_type attributes
		    } else {
			set_array_of_attributes attributes "" $class_c_info "" $return_type -1 -1 "cm" "r" {} 0
			analyze_method_pointers attributes
		    }
		    set_array_of_attributes attributes "" $class_c_info "" $list_of_arguments -1 -1 "cm" "a" {} 0
#call to a procedure for analyzing method arguments
		    process_list_of_arguments attributes
		}
	    }
	}
    }
}



#
# This procedure a list of start and end positions of every class with the 
# same name as my_class in the file called my_filename. 
#


proc extract_start_end_pos {my_class my_filename} {
    global db_classe
    global db_unions
    set class_start_end [$db_classe seq -col [list "0 =$my_class" 1 "2 =$my_filename" 3]]

# if there is more than one class with the same name as our class' in our file
    if {[llength $class_start_end] > 1} {
	set start_end_pos_list {}
	foreach element $class_start_end {
	    # make a list of all pairs (start position, end position)
	    lappend start_end_pos_list [list [lindex $element 1] [lindex $element 3]] 
	}
	return $start_end_pos_list
    } elseif {[llength $class_start_end] == 1} { ;# if there is only one class with this name in this file
	set class_start_end [lindex $class_start_end 0]
	#if we return just list, we will be returning a list with 2 elements. by returning list list we return a list of one element, which is a pair. 
	return [list [list [lindex $class_start_end 1] [lindex $class_start_end 3]]]
    } elseif {[llength $class_start_end] == 0} { ;# if there is no class with that name in our file
	if {[info exists db_unions]} { ;# maybe it's a union
	    set class_start_end [$db_unions seq -col [list "0 =$my_class" 1 "2 =$my_filename" 3]]
	    if {[llength $class_start_end] > 1} {
		set start_end_pos_list {}
		foreach element $class_start_end {
		    lappend start_end_pos_list [list [lindex $element 1] [lindex $element 3]] 
		}
		return $start_end_pos_list
	    } elseif {[llength $class_start_end] == 1} {
		set class_start_end [lindex $class_start_end 0]
		#if we return just list, we will be returning a list with 2 elements. by returning list list we return a list of one element, which is a pair. 
		return [list [list [lindex $class_start_end 1] [lindex $class_start_end 3]]]
	    }
	}
    }
    return {}
}

#
# This procedure is called by extract_method_return_and_argument_types. It checks whether 
# there is more than one argument in the list. If so, it calls analyze_templates_and_method_arguments.
# If it's only one argument, it calls find_real_type or analyze_method_pointers. 
#


proc process_list_of_arguments {attribute_list} {

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set list_args $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set global_argument $array_of_attributes(global_arg)

    if {$list_args != {}} {
	set index_of_comma [string first , $list_args]
	if {$index_of_comma != -1} { ;# if there's more than one argument
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $list_args -1 0 $coupling_type $method_coupling_type {} $global_argument
	    analyze_templates_and_method_arguments attributes
	} else { ;# if it's only one argument
	    set arg_type [extract_good_attribute_type $list_args]
	    if {$arg_type != 0} { ;# find its real type
		set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $arg_type -1 -1 $coupling_type $method_coupling_type {} $global_argument
		find_real_type attributes
	    } else { ;# it's a method pointer
	#use list_args instead of arg_type, which is 0.
		set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $list_args -1 -1 $coupling_type $method_coupling_type {} $global_argument
		analyze_method_pointers attributes
	    }
	}
    }
}

#original_class serves just for complex types, where we want to know both the class that contains our type, and the class that contains first part of the type, in order to well determine the type of second part


#
# This procedure tries to find the real type of some attribute (or method argument or return) type.
# It can be primitive type (no coupling), class type (coupling), template class type (coupling with 
# both template class and template arguments), complex type (A::B), typedef (which can be all of the mentioned) ... 
# This procedure calls appropriate procedures in order to find real type.
# 


proc find_real_type {attribute_list} {

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type_good $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)
  
#this may not be necessary now
    if {$my_class_info == ""} {
	set my_class_info $original_class_info
    }

# if the attribute type is a template, it will have "<>" in it. 
# Find the index of the first <.
    set index_of_less_than [string first < $attribute_type_good]
# if it's a template, set index, and call the appropriate procedure
    if {$index_of_less_than != -1} {
	set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $attribute_type_good -1 $index_of_less_than $coupling_type $method_coupling_type $t_argument $global_argument
	return [analyze_templates_and_method_arguments attributes]
    } else {
# if it's a complex type (A::B), find the index of ::
	set index_of_colon [string first :: $attribute_type_good]
# if it's a complex type, set the index of :: and call the appropriate procedure
	if {$index_of_colon != -1} {
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $attribute_type_good $index_of_colon -1 $coupling_type $method_coupling_type $t_argument $global_argument
	    return [check_complex_types attributes]
	} else { ;#if it's neither of previous cases, it must be a class or a typedef
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $attribute_type_good -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
	    return [check_classes_and_typedefs attributes]
	}
    }
}


#
# This procedure analyzes both templates and method arguments. If the index is 0,
# that means that it's method arguments and not template (index is the index of
# first <). 
# ....
#


proc analyze_templates_and_method_arguments {attribute_list} {

    global db_classe
    global parents_array
    global vars_c

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set index $array_of_attributes(index_of_less_than)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)

#added to solve NAME_NODE_PTR problem
    if {$my_class_info == ""} {
	set my_class_info $original_class_info
    }


#In order to use this proc to analyze method arguments, we have to think
#of attribute_type as template_args_string without <>. So if index = 0, 
#it means that we are analyzing method arguments : 

    if {$index != 0} {

	set template_class [string range $attribute_type 0 [incr index -1]]

#this is needed for complex types.
	set template_class_length [incr index 1]

# I don't think that it's necessary to check for classes, if it's a 
# template, it must be a class so ............ Not good, there are things like
# std::vector<> where std is a namespace (and SN doesn't treat namespaces). 
# Better look for it. 
	if {$template_class != [lindex $original_class_info 0] && $template_class != $inner_class && [lsearch [array names vars_c] $template_class] != -1 } {

#if template class is our class' parent, treat the coupling as 
#ancestor-descendant one.
	    set anc_desc_coupling 0
	    if {([info exists parents_array($original_class_info)]) && ([set parents $parents_array($original_class_info)] != {})} {
		foreach parent $parents {
		    set parent_class [lindex $parent 0]
		    if {$parent_class == $template_class} {
			set anc_desc_coupling 1
		    }
		}
	    }
	} else {
	    # if template class is not in .cl file we shouldn't treat its arguments, nor any complex type involving it
	    return 0
	}

# Usually template_args_string is everything after the name of the template class. So we'd just 
# get rid of the first < and the last >. But it's not that easy. There are more complex cases,
# such as :  A<B,C>::D or A<B,C>::D<..>, and we have to treat them all. So leave < and > for now.

	set template_args_string [string range $attribute_type $index end]
    } else { ;# if we are dealing with method arguments
	set template_args_string $attribute_type
    }

# if template arguments contain other templates, treat them properly

    if {[string first \> $template_args_string] != -1} {

# attempt to find indices of all < and > in the string that represents template
# arguments. put it in an array in the form (index >) or (index <)
	set string_part $template_args_string
	set start_index 0

	while {[string compare $string_part ""] !=0} {
	    regexp {[<>]} $string_part matched_expr 

#matched_expr is either < or >
	    regexp -indices {[<>]} $string_part index  ;#index is index of matched_expr : in fact pair of index of beg and end of an expression. both elements of the pair are the same numbers
	    set index [lrange $index 0 0] ;#extract only one element of the pair
	    set new_index [expr $start_index + $index] ;#we will discard the part before the matched symbol, in order to find another one in the next iteration. so keep track on the right index, that starts from the beginning of the initial string and not from the beginning of the kept part of the string

# switch sets the array explained above
	    switch $matched_expr {
		<  {set array_of_ltgt_indices($new_index) <}
		>  {set array_of_ltgt_indices($new_index) >}
	    }

#keep only the part of the initial string after the found symbol, in order to find the next one
	    set string_part [string range $string_part [expr $index + 1] end]

#set the start index to be the real index of the first element of the kept part
	    set start_index [expr $new_index + 1]

# reset index and matched_expr to avoid possible errors
	    set index 0
	    set matched_expr ""
	}

# now go through the array and find the matching < and > in order to keep them together 
	set count 0
	set start_index 0
	set index_of_paired_braces -1

	set to_continue 1

	while {$to_continue == 1} {
# array names command returns unsorted list. it has to be sorted in order to count <s and >s well
	    foreach ltgt_index [lsort -integer [array names array_of_ltgt_indices]] {
		if {$array_of_ltgt_indices($ltgt_index) == {<}} {
		    if {$count == 0} {
			set start_index $ltgt_index
		    }
		    set count [expr $count + 1]
		} elseif {$array_of_ltgt_indices($ltgt_index) == {>}} {
		    set count [expr $count - 1]
		}
		if {$count == 0} {
		    set index_of_paired_braces [expr $index_of_paired_braces + 1]
		    set array_of_paired_ltgt_indices($index_of_paired_braces) [list $start_index $ltgt_index]
		}
	    }
#if we have a simple case A<...>, than we'd like to remove first and last <>, and to check the template arguments that are left. So after 
#creating the array_of_paired_ltgt_indices check its first element : if the indices are 0 and template_args_string_length-1, it means that 
#we can remove them
	    set to_continue 0

	    if {[info exists array_of_paired_ltgt_indices(0)]} {
		if {[lindex $array_of_paired_ltgt_indices(0) 0] == 0} {
		    set template_args_string_length [string length $template_args_string]
		    set index_of_gt [lindex $array_of_paired_ltgt_indices(0) 1]
		    if {$index_of_gt == [expr $template_args_string_length - 1]} {
			unset array_of_ltgt_indices([lindex $array_of_paired_ltgt_indices(0) 0])
			unset array_of_ltgt_indices([lindex $array_of_paired_ltgt_indices(0) 1])
			unset array_of_paired_ltgt_indices(0)
			set remove_first_last_brace 1
			set count 0
			set start_index 0
			set index_of_paired_braces -1
			set to_continue 1
		    }
		}
	    }
	}
	
	set template_args_string_length [string length $template_args_string]

#added in order to try to solve a case like this : A<B::C,D>::E<F,G>
#necessary to keep initial <> (don't remove it, thinking of normal cases A<...>, because it will be wrong in a case A<...>::B<...>
#but if in fact we have a normal case, remove first and last <> and analyze the rest
#on the other hand, if it's a complex case, treat it as such

#NB I'm not certain why I do this checking again, and why I don't just use this code, and erase the similar one above. If I have time I'll check it. 	
	if {[info exists array_of_paired_ltgt_indices(0)]} {
	    if {[lindex $array_of_paired_ltgt_indices(0) 0] == 0} {
		set template_args_string_length [string length $template_args_string]
		set index_of_gt [lindex $array_of_paired_ltgt_indices(0) 1]
		if {$index_of_gt == [expr $template_args_string_length - 1]} {
		    unset array_of_paired_ltgt_indices(0)
		} else {
		    if {[string range $template_args_string [expr $index_of_gt + 1] [expr $index_of_gt + 2]] == "::" } {
			set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $attribute_type [expr $index_of_gt + [expr $template_class_length + 1]] -1 $coupling_type $method_coupling_type $t_argument $global_argument
			check_complex_types attributes
			return 0 ;#check this
		    }
		}
	    }
	}


	if {[info exists template_class]} {
	    #puts "Coupling between [lindex $original_class_info 0] and template_class : $template_class"
	    set_coupling_array $original_class_info $template_class $coupling_type $method_coupling_type $anc_desc_coupling
	}

#if there are templates inside of template_args_string
	if {[array size array_of_paired_ltgt_indices] > 0} {
# change all commas for | in the parts of the string of template args between < and > delimited by indices in the array_of_paired_ltgt_indices

	    foreach {index {list_of_indices}} [array get array_of_paired_ltgt_indices] {
		regsub -all {,} [string range $template_args_string [lindex $list_of_indices 0] [lindex $list_of_indices 1]] {|} array_of_modified_parts($list_of_indices)
	    }

# after that re-create template args with modified strings in order to separate it afterwards using comma as a separator

	    set previous_index 0
	    foreach {index list_of_indices} [array get array_of_paired_ltgt_indices] {
		append template_args_with_protected_commas [string range $template_args_string $previous_index [expr [lindex $list_of_indices 0] - 1]]
		append template_args_with_protected_commas $array_of_modified_parts($list_of_indices)
		set previous_index [expr [lindex $list_of_indices 1] + 1]
	    }
	    if {$previous_index != [string length $template_args_string]} {
		append template_args_with_protected_commas [string range $template_args_string $previous_index end]
	    }
#we deal here with the whole template arg string, where we kept outer <>; we can now get rid of them
	    if {[info exists remove_first_last_brace]} {
		set end_index [expr [string length $template_args_with_protected_commas] - 2]
		set template_args_with_protected_commas [string range $template_args_with_protected_commas 1 $end_index]
	    }
	} else { ;#if there is no templates inside template arguments string
#remove <>
	    set end_index [expr [string length $template_args_string] - 2]
	    set template_args_string [string range $template_args_string 1 $end_index]
	    set template_args_with_protected_commas $template_args_string
	}
    } else { ;#end of if template args contain other templates
	set template_args_with_protected_commas $template_args_string
    }
# now that the commas inside <> are protected, we can separate the list of template arguments on the actual arguments and remove the comma protection

    set template_args_list [split $template_args_with_protected_commas ,]

    set arg_position -1
    set return_value {}
    foreach element $template_args_list {
	set arg_position [incr arg_position 1]
	if {[string first \< $element] != -1} {
	    regsub -all {\|} $element {,} element
	}
	if {[info exists template_class]} {
	    set t_argument [list $template_class $arg_position]
	} else {
	    set t_argument {}
	}
	set arg_type [extract_good_attribute_type $element]
	if {$arg_type !=0} {
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $arg_type -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
	    find_real_type attributes
	    if {[info exists template_class]} {
		set class_file [$db_classe seq -col [list "0 =$template_class" 2]]
		set return_value [lappend return_value [lindex $class_file 0]]
	    }
	} else {
#use element instead of arg_type
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $element -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
	    analyze_method_pointers attributes
	}
    }

    return [eliminate_duplicates_from_list $return_value]
}


proc check_classes_and_typedefs {attribute_list} {

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)

    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $attribute_type -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
    set return_from_check_classes [check_type_for_classes attributes]

    if {$return_from_check_classes == 0} {
	return [check_type_for_typedefs attributes]
    }
#added for complex types :
    return $return_from_check_classes
}


# for types that have :: we have to check both sides. this procedure does it. 

proc check_complex_types {attribute_list} {

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set index $array_of_attributes(index_of_colon)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)

    if {$my_class_info == ""} {
	set my_class_info $original_class_info
    }

    set first_part [string range $attribute_type 0 [expr $index - 1]]
    set second_part [string range $attribute_type [expr $index + 2] end]

#just in case we have a pointer or something similar : *T::X 
    set first_part [extract_good_attribute_type $first_part]
#changed for complex types
    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $first_part -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
    set return_from_classes_typedef [check_classes_and_typedefs attributes]

    #here it was || before, but it wouln't work for -1,s so I put &&. hope it works 

    if {($return_from_classes_typedef != 0) && ($return_from_classes_typedef != -1)} {

#in case we have T::* (usual case with method pointers	

	if {[llength $return_from_classes_typedef] == 1} {
	    set second_part [extract_good_attribute_type $second_part]
	    set return_from_classes_typedef [lindex $return_from_classes_typedef 0]
	    set my_class [lindex $return_from_classes_typedef 0]
	    set my_file [lindex $return_from_classes_typedef 1]
	    set start_end_pos [lindex [extract_start_end_pos $my_class $my_file] 0]
#length of start_end_pos is always 1
	    set my_class_info [list $my_class [lindex $start_end_pos 0] [lindex $start_end_pos 1] $my_file]
	    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $second_part -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
	    check_classes_and_typedefs attributes
	}
    }
}



#
#  Attribute types can be in the following form : const NameOfTheClass, or 
#  int NameOfTheClass *, or unsigned int, or int A::B * ... 
#  This procedure extract the right (class) parameter.  
#
#
proc extract_good_attribute_type {attribute_type} {


# if attribute_type contains '(', it means that it's a function type
# treat it as class - method coupling.

    foreach attribute_type_part $attribute_type {
	if {[string first ( $attribute_type_part] != -1} {
	    #puts "treat as class-method coupling $attribute_type"
	    #add return 0 here. check_type_for_typedefs will use it to treat method-pointers.
	    return 0
	}
    }

#erase unnecessary types : substitute "int ", "unsigned ", "const ", "char " ... by nothing 
# N.B. don't ever forget braces around \4 

#Changed (\ &) to ((\ )*&)in order to solve problem with Binding
# added inline and typename for return types and template args respectively
#added space (\ ) in a-z part to solve scaled1D<typename TwoD...> 

    regsub {(int\ |const\ |char\ |unsigned\ |virtual\ |inline\ |typename\ )*([a-zA-Z0-9,<>_:\*\ ]+)(((\ (\*)*)|((\ )*&)|(\ (\*)*\[\]))*)} $attribute_type {\2} attribute_type_good
#    regsub -all {([^int][^unsigned][^const][^char])*(int|unsigned|const|char)(\ )*([ ])*} $attribute_type {\1\4} attribute_type

#if there is a macro before 
    if {[lsearch [array names vars_ma] $attribute_type_good] != -1} {
	set attribute_type_good [lindex $attribute_type 1]
    }

#remove * attached to the attribute name : case in template arguments. remove spaces before * too
    regsub {([^\*]*)(\*)*} $attribute_type_good {\1} attribute_type_good 

#try to eliminate just the last space  
    return [string trimright $attribute_type_good]
}



proc eliminate_duplicates_from_list {list} {
    foreach element $list {
	set array_without_duplicates($element) 1
    }
    return [lsort [array names array_without_duplicates]]
}


#
#  This procedure checks whether the attribute type can be found in
#  the array of class names. If so, there is coupling. If not, return 0.
#


proc check_type_for_classes {attribute_list} {

    global vars_c
    global db_classe
    global templates
    global template_arg_coupling
    global parents_array

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)


    if {$my_class_info == "" || [lindex $my_class_info 0] == {} } {
	set my_class_info $original_class_info
    }

    set filename [lindex $my_class_info 3]

# in order to avoid conflict between typedef names and template arguments' 
# names we should check first if the class is a template, and if the attribute
# is of the template's attribute type. if not we check if the attribute is of
# some other class' type. 

#change original_class for my_class

    set original_class [lindex $original_class_info 0]
    set my_class [lindex $my_class_info 0]

    if {[is_template $original_class]} {
	if {[lsearch $vars_c($original_class) $attribute_type] != -1} {
	    #puts "Coupling between $original_class and its template argument $attribute_type"

#FIXIT : see whether method_coupling_type changes something in the array - eliminate duplicates
	    set template_arg_coupling($original_class_info) [lappend template_arg_coupling($original_class_info) [list $attribute_type $coupling_type $method_coupling_type]]
	    return ;#if we don't return here it might find that the template type is also a typedef somewhere else, and it would be erroneous coupling
	}
    }

# if the attribute_type is listed in the class list, there is coupling 
# between these two classes 

    if {$attribute_type != $original_class && $attribute_type != $inner_class} {

	if {[lsearch [array names vars_c] $attribute_type] != -1} {

#check whether attribute_type is one of inner classes of original_class. if yes then there is no coupling
	    foreach element [extract_start_end_pos $attribute_type $filename] {
		set start_end_line_at [extract_line_numbers [list $element {}]]
		set start_line_at [lindex $start_end_line_at 0]
		set end_line_at [lindex $start_end_line_at 1]
		set outer_class_info [is_inner_class $start_line_at $end_line_at $filename]
		if {$outer_class_info != 0} {
		    set outer_class [lindex $outer_class_info 0]
#there are 2 cases : the second is B::X, where X is inner class of B, and therefore there should not be any coupling
		    if {$outer_class == $original_class || $outer_class == $my_class} {
			return -1 ;#see what to return : it cannot be 0 because it shouldn't continue further : we found the type but it's the same
		    }
		}
	    }

#for templates : SEE WHAT TO DO IF ATTRIBUTE_TYPE is inner class of ORIGINAL_CLASS - no coupling, but still have to update templates array
	
	    if {$t_argument != ""} {
		set template_class [lindex $t_argument 0]
		set arg_position [lindex $t_argument 1]
		set templates($template_class,$arg_position) [lappend templates($template_class,$arg_position) $attribute_type]
	    }
	    set anc_desc_coupling 0
	    if {([info exists parents_array($original_class_info)]) && ([set parents $parents_array($original_class_info)] != {}) } {
		foreach parent $parents {
		    set parent_class [lindex $parent 0]
		    if {$parent_class == $attribute_type} {
			set anc_desc_coupling 1
		    }
		}
	    }	    

	    #puts "Coupling between $original_class and $attribute_type"
	    set_coupling_array $original_class_info $attribute_type $coupling_type $method_coupling_type $anc_desc_coupling

	    #added to solve complex types problem : return file to which that class belongs
	    set class_file [$db_classe seq -col [list "0 =$attribute_type" 2]]

#try to return both the class name and the filename, because now we have to make sure if the typedef is not only a part of the same file, but of the same class as well : see check_type_for_typedefs
	    return [list [lindex $class_file 0]]

	}
	return 0 ;#if att. type doesn't have the same name as my_class and it's not a class
    }

    return -1 ;#type has the same name as my class - NO COUPLING (not sure whether -1 is a good return value)
}
#    return 0 - if we return here than in the case of the class Binding having as a method parameter class Binding, it will try to find a typedef with a name Binding, and it will find it (wrong).




#
#  This procedure checks whether the attribute type is in the list  
#  of typedef names. If it is, it checks corresponding typedefs type
#  (from the list of typedefs types) for classes. If it's a class,
#  there will be coupling. If not, it calls recursively this procedure to
#  check whether it's a typedef. 
#


proc check_type_for_typedefs {attribute_list} {
    global array_included_files
    global db_typedefs
    global parents_array

    if {![info exists db_typedefs]} {
	return 0
    }

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)
    set global_argument $array_of_attributes(global_arg)

    if {$my_class_info == ""} {
	set my_class_info $original_class_info
    }

    set global_list {}
    set global_in_included_files {}

#set class start and end line
    set start_line [lindex $my_class_info 1]
    set start_line [lindex [extract_line_numbers [list $start_line {}]] 0]

    set end_line [lindex $my_class_info 2]
    set end_line [lindex [extract_line_numbers [list $end_line {}]] 0]

# I put my_class_info for global typedefs. CHECKIT

    set filename [lindex $my_class_info 3]

#extract all the typedefs with the same name from the typedef database

    set td_type [$db_typedefs seq -col [list "0 =$attribute_type" 1 2 5]]

#I have to return something from this procedure : needed for complex types (in case of A::B, where A is a typedef). if I return it in the middle of foreach, I won't be able to see other typedefs with the same name. So I'm going to collect return values in this variable.
    set return_value {}

    if {[llength $td_type] != 0} {
	#if it's found in our class or parent's class, this flag serves to tell me not to continue
	set found 0

	foreach element $td_type {
	    set parents_attributes {}
	    set line_no [extract_line_numbers [list [lindex $element 1] {}]]
	    set line_no [lindex $line_no 0]
	    
#extract type of the typedef
	    set type [lindex $element 3]
#NB : type_of_typedef can be 0 if it's a method-pointer

	    set type_of_typedef [extract_good_attribute_type $type]
#extract filename

	    set typedef_filename [lindex $element 2]  

#if typedef is not global (this is not the first pass through this procedure : if we first found that our typedef is global and now we're looking for its type, we will surely not find it in our class or parent class: it must be global again. so skip this part and go directly to the global. 
	    if {$global_argument == 0} {
	    #if the typedef is in our file, and in our class
		if {($typedef_filename == $filename) && [is_inside_my_class $filename $start_line $end_line $line_no]} {

#!!!!!HERE YOU HAVE TO INTRODUCE NOT ONLY START_LINE AND END_LINE, BUT AN ARRAY OF LINE_NUMBERS, IF YOU HAVE INNER_CLASSES (MAKE A METHOD FOR IT, IF IT'S NOT ALREADY THERE). YOU ARE NOT INTERESTED IN TYPEDEFS DEFINED IN INNER CLASSES, IT WOULD BE AN ERROR TO INTRODUCE THEM.

		    if {[string compare $type_of_typedef ""] != 0} {
			set found 1
			#added for method-pointers
			if {$type_of_typedef != 0} {
			    if {$type_of_typedef != $attribute_type} {
				set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $type_of_typedef -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
				#added for complex types	
				#this is because check_type_for_classes returns list [lindex ...]
				set return_value [concat $return_value [find_real_type attributes]]
			    }
			} else {
			    #use type instead of typedef
			    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $type -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
			    analyze_method_pointers attributes
			    set return_value 0
			}
		    }    		
		} elseif {([info exists parents_array($my_class_info)]) && ([set parents $parents_array($my_class_info)] != {})} { ;#if our class has parents
		    foreach parent $parents {
			set parent_start_pos [lindex $parent 1]
			set parent_start_line [lindex [extract_line_numbers [list $parent_start_pos {}]] 0]
			set parent_end_pos [lindex $parent 3]
			set parent_end_line [lindex [extract_line_numbers [list $parent_end_pos {}]] 0]
			set parent_filename [lindex $parent 2]
			set parent_class [lindex $parent 0]
			
			if {($parent_filename == $typedef_filename) && ($line_no > $parent_start_line && $line_no < $parent_end_line)} {			    ##CHANGE IF YOU CAN
			    #attention : carefully change this procedure's parameters
#if I kept parent_start/end_line, it produces wrong results. Have to keep position. 
			    set par_start_end_pos [list $parent_start_pos $parent_end_pos]
			    #type at the end added for method-pointers
			    lappend parents_attributes [list $parent_class $par_start_end_pos $type_of_typedef $parent_filename $type]
			}
		    }
		}
	    }
	    
	    if {($found == 0) || ($global_argument == 1)} {
		#is it original or my_class ?
		set filename [lindex $my_class_info 3]
		
		if {($typedef_filename == $filename) && ([is_global $type $line_no $filename])} {
		    lappend global_list [list $type_of_typedef $type] ;#type added for method-pointers
		} elseif {[info exists array_included_files($filename)]} { ;#if our file includes other .h
		    foreach entry_if $array_included_files($filename) {
			#for every such file, see if it contains our typedef
			if {[string compare $entry_if $typedef_filename] == 0} {
			    # if it is global in an included file
			    if {[is_global $type $line_no $typedef_filename]} {
				if {[string compare $type_of_typedef ""] != 0} {
				    lappend global_in_included_files [list $type_of_typedef $typedef_filename $type] ;#type added for method-pointers
				}
			    }
			}
		    }
		}
	    }
	    
	}
	
    
	
	
#NB : if we check every possible typedef using a list, it is possible that we find first td in parent's class, and then in our class. FIX IT !!!!

	
	if {($found == 0) && ([llength $parents_attributes] > 0)} {
	    set original_class [lindex $original_class_info 0]
	    foreach parent_att $parents_attributes {
		set type_of_typedef [lindex $parent_att 2] ;#added for method-pointers
		set parent_class [lindex $parent_att 0]
		#IF YOU CHANGE PARENTS_ATTRIBUTES, CHANGE THIS ALSO
		set parent_class_info [list $parent_class [lindex $parent_att 1] [lindex $parent_att 3]]
		if {$type_of_typedef != 0} {
		    if {$type_of_typedef != $attribute_type} {
			set_array_of_attributes attributes $inner_class $original_class_info $parent_class_info [lindex $parent_att 2] -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
			#puts "Coupling between $original_class and parent_class $parent_class because of typedef $type_of_typedef"
			
			#here you can maybe put parent_class_info ?
			set_coupling_array $original_class_info $parent_class $coupling_type $method_coupling_type 1
			
			set return_value [concat $return_value [find_real_type attributes]]
		    }
		} else {
		    set_array_of_attributes attributes $inner_class $original_class_info $parent_class_info [lindex $parent_att 4] -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
		    #puts "Coupling between $original_class and parent_class $parent_class because of typedef [lindex $parent_att 4]"
		    #here you can maybe put parent_class_info ?
		    set_coupling_array $original_class_info $parent_class $coupling_type $method_coupling_type 1
		    analyze_method_pointers attributes
		    set return_value 0
		}
	    }
	}

	if {([llength $global_list] != 0) && ($found == 0)} {
	    foreach global_td $global_list {
		set type_of_typedef [lindex $global_td 0]
		if {$type_of_typedef != 0} {
		    if {$type_of_typedef != $attribute_type} {
			#change global_argument to 1
			set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $type_of_typedef -1 -1 $coupling_type $method_coupling_type $t_argument 1
			set return_value [concat $return_value [find_real_type attributes]]
		    }
		} else {
		    #use type instead of type_of_typedef
		    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info [lindex $global_td 1] -1 -1 $coupling_type $method_coupling_type $t_argument 1
		    analyze_method_pointers attributes
		    set return_value 0
		}
	    }  
	} elseif {([llength $global_in_included_files] != 0) && ($found == 0)} {
	    foreach global_td $global_in_included_files {
		set type_of_typedef [lindex $global_td 0]
		set my_class_info [list "" "" "" [lindex $global_td 1]]
		if {$type_of_typedef != 0} {
		    if {$type_of_typedef != $attribute_type} {
			set_array_of_attributes attributes $inner_class $original_class_info $my_class_info [lindex $global_td 0] -1 -1 $coupling_type $method_coupling_type $t_argument 1
			set return_value [concat $return_value [find_real_type attributes]]
		    }
		} else {
		    set_array_of_attributes attributes $inner_class $original_class_info $my_class_info [lindex $global_td 2] -1 -1 $coupling_type $method_coupling_type $t_argument 1
		    analyze_method_pointers attributes
		    set return_value 0
		}
	    }  
	}
    }
	
    
    if {($return_value != 0) && ([llength $return_value] > 1)} {
	return [eliminate_duplicates_from_list $return_value]
    }
    return $return_value
    
}



proc analyze_method_pointers {attribute_list} {

    upvar $attribute_list array_of_attributes
    set inner_class $array_of_attributes(inner_class)
    set original_class_info $array_of_attributes(original_class_info)
    set my_class_info $array_of_attributes(class_c_info)
    set attribute_type $array_of_attributes(attribute_type_good)
    set coupling_type $array_of_attributes(coupling_type)
    set method_coupling_type $array_of_attributes(method_coupling_type)
    set t_argument $array_of_attributes(t_argument)   
    set global_argument $array_of_attributes(global_arg)
    
    set method_args_list [split $attribute_type (]
    
    foreach element $method_args_list {
	set type [extract_good_attribute_type [string trim $element )]]
	set_array_of_attributes attributes $inner_class $original_class_info $my_class_info $type -1 -1 $coupling_type $method_coupling_type $t_argument $global_argument
	find_real_type attributes
    }
}




#this procedure has to be changed to accept start/end pos and not start/end line
proc has_parents {my_class start_line end_line filename correct_parents_list index} {
    global db_classe
    global db_inheritance
    global array_included_files
    global db_typedefs

    set parents [$db_inheritance seq -col [list "0 =$my_class" 1 2 "3 =$filename" 4 5 8]]
# 8 can serve us for template arguments, I'm not sure

    set parents_list {}
    set parent_class_in_my_file {}
    set parent_class_in_included_file {}

    foreach element $parents {
	set line [lindex $element 2]
	set parent_line_no [lindex [extract_line_numbers [list $line {}]] 0]
	if {$parent_line_no >= $start_line && $parent_line_no <= $end_line} {
	    set parent_class [lindex $element 1]
	    lappend parents_list $parent_class
	}
    }

#we have to return only parent classes that are in our file or included file, to avoid possible errors if we have two classes with the same name, but in different files

    if {$parents_list != {}} {
	foreach parent $parents_list {
#if our file includes other .h
	    set parent_class_in_my_file [$db_classe seq -col [list "0 =$parent" 1 "2 =$filename" 3]]
	    if {$parent_class_in_my_file != {}} {
		set correct_parents_list [concat $correct_parents_list $parent_class_in_my_file]
	    } elseif {[info exists array_included_files($filename)]} { 
		foreach entry_if $array_included_files($filename) { 
		    set parent_class_in_included_file [$db_classe seq -col [list "0 =$parent" 1 "2 =$entry_if" 3]]
		    if {$parent_class_in_included_file != {}} {
			set correct_parents_list [concat $correct_parents_list $parent_class_in_included_file]
		    }
		} 
	    } 

	    if {($parent_class_in_my_file == {}) && ($parent_class_in_included_file == {}) && ([info exists db_typedefs]) } {
		#if parent class is a typedef (ACE_Select_Reactor), and it is defined in my file 
		set parent_typedef_in_my_file [$db_typedefs seq -col [list "0 =$parent" 1 "2 =$filename" 5]]
		if {$parent_typedef_in_my_file != {}} {
		    set parent_type [lindex $parent_typedef_in_my_file 4]
		    set parent_class [$db_classe seq -col [list "0 =$parent_type" 1 2 3]]
		    #if the typedef is a class
		    if {$parent_class != {}} {
			set correct_parents_list [concat $correct_parents_list $parent_class]
		    } else {
			#if the typedef is another typedef
			set parent_typedef [$db_typedefs seq -col [list "0 =$parent_type" 1 2 5]]
			if {$parent_typedef != {}} {
			    set correct_parents_list [concat $correct_parents_list $parent_typedef]
			}
		    }
		} elseif {[info exists array_included_files($filename)]} { 
		    foreach entry_if $array_included_files($filename) { 
			set list_parent_typedef_in_included_file [$db_typedefs seq -col [list "0 =$parent" 1 "2 =$entry_if" 3 5]]
			if {$list_parent_typedef_in_included_file != {}} {
			    foreach parent_typedef_in_included_file $list_parent_typedef_in_included_file {
				set parent_type [lindex $parent_typedef_in_included_file 4]
				set parent_class [$db_classe seq -col [list "0 =$parent_type" 1 2 3]]
				if  {$parent_class == {}} {
				    #parent_type can be a template : extract the class name of the template. But don't extract it too soon, because some classes are kept with template arguments in .cl database. Now is fine.
				    regexp {([^<]+)} $parent_type parent_type 				    
				    set parent_class [$db_classe seq -col [list "0 =$parent_type" 1 2 3]]
				}
				#if the typedef is a class
				if {$parent_class != {}} {
				    set correct_parents_list [concat $correct_parents_list $parent_class]
				} else {
				    #if the typedef is another typedef
				    set parent_typedef [$db_typedefs seq -col [list "0 =$parent_type" 1 2 3 5]]
				    if {$parent_typedef != {}} {
					set correct_parents_list [concat $correct_parents_list $parent_typedef]
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    }

    if {$index < [llength $correct_parents_list]} {
	set parent [lindex $correct_parents_list $index]
	set parent_class [lindex $parent 0]
	set parent_start_line [lindex $parent 1]
	set parent_start_line [lindex [extract_line_numbers [list $parent_start_line {}]] 0]
	set parent_end_line [lindex $parent 3]
	set parent_end_line [lindex [extract_line_numbers [list $parent_end_line {}]] 0]
	set parent_filename [lindex $parent 2]
	has_parents $parent_class $parent_start_line $parent_end_line $parent_filename $correct_parents_list [expr $index + 1]
    } else {
	return $correct_parents_list
    }

}

proc is_global {type line_no filename} {
    global db_classe
    set classes_line_nos [$db_classe seq -col [list 1 "2 =$filename" 3]]
    if {$classes_line_nos != {}} {
	set start 0
	foreach element [lsort $classes_line_nos] {
	    set end [lindex $element 0]
	    if {$line_no > $start && $line_no < $end} {
		return 1 ;#is global, is found between classes in that .h
	    }
	    set start [lindex $element 2]
	}
    } else {
	return 1 ;# this means that this particular .h doesn't have class declarations in it
    }
    return 0
}

proc extract_line_numbers {list_start_end_pos} {

#we have to do this, because we have a list of only one element, a pair. 
    set list_start_end_pos [lindex $list_start_end_pos 0]

    set start_pos [lindex $list_start_end_pos 0]
    set index_of_dot [string first . $start_pos]
    set start_pos [string range $start_pos 0 [incr index_of_dot -1]]
    set start_pos [string trimleft $start_pos 0]
    set end_pos [lindex $list_start_end_pos 1]
   #this procedure is sometimes used only with one argument, so cut the long story short
    if {$end_pos == ""} {
	set end_pos {}
    } else {
	set index_of_dot [string first . $end_pos] 
	set end_pos [string range $end_pos 0 [incr index_of_dot -1]]
    }
    return [list $start_pos $end_pos]
}


proc finish_template_arg_coupling {} {
    global template_arg_coupling
    global templates

    if {(![info exists template_arg_coupling])||(![info exists templates])} {
	setLog "No templates in the project."
	return
    }


    foreach {coupled_template_class_pos_filename coupled_template_argument_list} [array get template_arg_coupling] {
	set coupled_template_class [lindex $coupled_template_class_pos_filename 0]
	set coupled_template_class_start_line [lindex $coupled_template_class_pos_filename 1]
	set coupled_template_class_end_line [lindex $coupled_template_class_pos_filename 2]
	set coupled_template_class_filename [lindex $coupled_template_class_pos_filename 3]
	foreach {template_class_and_position template_arg_types} [array get templates] {
	    set template_class [lindex [split $template_class_and_position ,] 0]
	    set arg_name [lindex $template_arg_types 0]
	    regsub {(class\ )*([a-zA-Z0-9_:]+)} $arg_name {\2} arg_name
	    set list_template_types [eliminate_duplicates_from_list [lrange $template_arg_types 1 end]]
	    

	    foreach coupled_template_argument_coupl_type $coupled_template_argument_list {
		set coupled_template_argument [lindex $coupled_template_argument_coupl_type 0]
		set coupling_type [lindex $coupled_template_argument_coupl_type 1]
		set method_coupling_type [lindex $coupled_template_argument_coupl_type 2]
		
		if {$coupled_template_class == $template_class && $coupled_template_argument == $arg_name} {
	    
		    foreach arg_type $list_template_types {
			#puts "Coupling between $coupled_template_class and $arg_type"
			set anc_desc_coupling 0
			if {([info exists parents_array($coupled_template_class_pos_filename)])  && ([set parents $parents_array($coupled_template_class_pos_filename)] != {}) } {
			    #if parent is template argument
			    foreach parent $parents {
				set parent_class [lindex $parent 0]
				if {$parent_class == $coupled_template_argument} {
				    set anc_desc_coupling 1
				}
			    }
			}
			set_coupling_array $coupled_template_class_pos_filename $arg_type $coupling_type $method_coupling_type $anc_desc_coupling
		    }
					     	    
#FIX THIS : no list start end, but real start_end_pos. if you don't fix this you will have duplicates or errors. 
			
			#set_coupling_array $coupled_template_class_pos_filename $arg_type $coupling_type $method_coupling_type $anc_desc_coupling
		}
	    }
	}
    }
}



#This procedure sets coupling arrays. For example :
# for import_ca_coupling_array : array element name is $my_class and value is a list of classes used by that class as attributes
# for export_ca_coupling_array : array element name is &attribute_type and value is a list of all the classes that use that type as attribute

#As for the arguments I added class start_end_pos and class_filename, but I didn't use it because I know with certainty only line_numbers and filename for coupled_class and not for coupled_attribute, and I would need both, if I want to make arrays like it's explained above.  


proc set_coupling_array {coupled_class_info coupled_attribute coupling_type method_coupling_type ancestor_descendant_coupling} {
    global db_classe
    global db_unions

    global import_ca_coupling_array
    global export_ca_coupling_array
    global import_cm_coupling_array
    global export_cm_coupling_array
    global ancestor_import_ca_coupling_array
    global descendant_export_ca_coupling_array
    global ancestor_import_cm_coupling_array
    global descendant_export_cm_coupling_array
    global array_included_files

    set coupled_attribute_list {}
    set coupled_attribute_list_included {}

    #coupled_attribute_info contains attribute's class or union, start and end position and filename
    set coupled_attribute_info [$db_classe seq -col [list "0 =$coupled_attribute" 1 3 2]]

    if {$coupled_attribute_info == {}} {
	if {[info exists db_unions]} {
	    set coupled_attribute_info [$db_unions seq -col [list "0 =$coupled_attribute" 1 3 2]]
	}
    }


    foreach coupled_attribute_class_info $coupled_attribute_info {
	
	if {$coupling_type == "ca"} {
	    if {$ancestor_descendant_coupling} {
		set ancestor_import_ca_coupling_array($coupled_class_info) [lappend ancestor_import_ca_coupling_array($coupled_class_info) $coupled_attribute_class_info]
		set descendant_export_ca_coupling_array($coupled_attribute_class_info) [lappend descendant_export_ca_coupling_array($coupled_attribute_class_info) $coupled_class_info]
	    } else {
		set import_ca_coupling_array($coupled_class_info) [lappend import_ca_coupling_array($coupled_class_info) $coupled_attribute_class_info]
		set export_ca_coupling_array($coupled_attribute_class_info) [lappend export_ca_coupling_array($coupled_attribute_class_info) $coupled_class_info]
	    }
	} elseif {$coupling_type == "cm"} {
	    if {$ancestor_descendant_coupling} {
		set ancestor_import_cm_coupling_array($coupled_class_info) [lappend ancestor_import_cm_coupling_array($coupled_class_info) $coupled_attribute_class_info]
		set descendant_export_cm_coupling_array($coupled_attribute_class_info) [lappend descendant_export_cm_coupling_array($coupled_attribute_class_info) $coupled_class_info]
	    } else {
# if I keep method_coupling_type, then it treats differently same class with different method_coupling_type (important for elimination of duplicates, in order to calculate a number of coupled classes and not instances)
		set import_cm_coupling_array($coupled_class_info) [lappend import_cm_coupling_array($coupled_class_info) $coupled_attribute_class_info]
		#set import_cm_coupling_array($coupled_class_info) [lappend import_cm_coupling_array($coupled_class_info) [list $coupled_attribute_class_info $method_coupling_type]]
		set export_cm_coupling_array($coupled_attribute_class_info) [lappend export_cm_coupling_array($coupled_attribute_class_info) $coupled_class_info]
		#set export_cm_coupling_array($coupled_attribute_class_info) [lappend export_cm_coupling_array($coupled_attribute_class_info) [list $coupled_class_info $method_coupling_type]]
	    }
	}	
    }
}


#this procedure quantifies coupling. It's easy to quantify : it's in fact length of each element of the coupling array (which is in fact a list of coupled classes). This procedure creates also an output file.



proc quantify_coupling {path projname access_option} {
    global import_ca_coupling_array
    global export_ca_coupling_array
    global import_cm_coupling_array
    global export_cm_coupling_array
    global ancestor_import_ca_coupling_array
    global descendant_export_ca_coupling_array
    global ancestor_import_cm_coupling_array
    global descendant_export_cm_coupling_array   
    global DIT_array

 #   global templates

   
    #puts "**** EXPORT_CM_COUPLING ****"

    #foreach element [lsort [array names export_cm_coupling_array]] {
#	puts [join [list $element $export_cm_coupling_array($element) [llength $export_cm_coupling_array($element)]] \t]
#    }
#
#    puts "**** IMPORT_CM_COUPLING ****"
#
#    foreach element [lsort [array names import_cm_coupling_array]] {
#	puts [join [list $element $import_cm_coupling_array($element) [llength $import_cm_coupling_array($element)]] \t]
#    }
#
#    puts "**** EXPORT_CA_COUPLING ****"
#
#     foreach element [lsort [array names export_ca_coupling_array]] {
#	puts [join [list $element $export_ca_coupling_array($element) [llength $export_ca_coupling_array($element)]] \t]
#    }
#
#    puts "**** IMPORT_CA_COUPLING ****"
#
#     foreach element [lsort [array names import_ca_coupling_array]] {
#	puts [join [list $element $import_ca_coupling_array($element) [llength $import_ca_coupling_array($element)]] \t]
#    }
#
    #puts "**** TEMPLATES ****"

    #foreach element [lsort [array names templates]] {
	#set template_arg_name [lindex $templates($element) 0]
	#set list_template_types [lrange $templates($element) 1 end] 
	#puts [join [list $element [list $template_arg_name] [list [eliminate_duplicates_from_list $list_template_types]]] \t]
    #}
#

#in order to list every coupled class, we concatanate all array names, elliminate duplicates and sort them


    set all_coupled_elements [concat [array names export_cm_coupling_array] [array names import_cm_coupling_array] [array names export_ca_coupling_array] [array names import_ca_coupling_array] [array names descendant_export_cm_coupling_array] [array names ancestor_import_cm_coupling_array] [array names descendant_export_ca_coupling_array] [array names ancestor_import_ca_coupling_array]]
    set all_coupled_elements [lsort [eliminate_duplicates_from_list $all_coupled_elements]]

    
#if there is coupling put the length of coupling list; if not, put 0
    
#set couplingFileId [open $path/$projname-[string trim $access_option -]-coupling.csv w 0600]
    if {[catch {open $path/$projname-[string trim $access_option -]-coupling.csv w 0600} couplingFileId]} {
	setLog "Unable to open output file $path/$projname-[string trim $access_option -]-coupling.csv"
	return
    }

#changed \t by comma, in order to have standard csv file
  
    puts $couplingFileId "Class Name,Start position,End position,Filename,CAIC (instances),CAIC (classes),CAEC (instances),CAEC (classes),CMIC (instances),CMIC (classes),CMEC (instances),CMEC (classes),ACAIC (instances),ACAIC (classes),DCAEC (instances),DCAEC (classes),ACMIC (instances),ACMIC (classes),DCMEC (instances),DCMEC (classes),LOC,Number of [string trim $access_option -] attributes,Number of [string trim $access_option -] methods,DIT"

puts "Creating metrics file ... (please wait)"

    foreach element $all_coupled_elements {
#puts -nonewline "."
	puts -nonewline $couplingFileId "[lindex $element 0],[lindex $element 1], [lindex $element 2],[lindex $element 3],"
	if {[info exists import_ca_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $import_ca_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $import_ca_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists export_ca_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $export_ca_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $export_ca_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists import_cm_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $import_cm_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $import_cm_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists export_cm_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $export_cm_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $export_cm_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists ancestor_import_ca_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $ancestor_import_ca_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $ancestor_import_ca_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists descendant_export_ca_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $descendant_export_ca_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $descendant_export_ca_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists ancestor_import_cm_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $ancestor_import_cm_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $ancestor_import_cm_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
	if {[info exists descendant_export_cm_coupling_array($element)]} {
	    puts -nonewline $couplingFileId "[llength $descendant_export_cm_coupling_array($element)],"
	    puts -nonewline $couplingFileId "[llength [eliminate_duplicates_from_list $descendant_export_cm_coupling_array($element)]],"
	} else {
	    puts -nonewline $couplingFileId "0,"
	    puts -nonewline $couplingFileId "0,"
	}
#put LOC, number of pub/prot/priv/all attr, number of pub/prot/priv/all methods
	puts -nonewline $couplingFileId [extract_loc_na_nm $element $access_option]
#put DIT
#here I don't use lilne_numbers and filename, it's easier this way. see if it's completely wrong or not. 
	if {[info exists DIT_array([lindex $element 0])]} {
	    puts -nonewline $couplingFileId "$DIT_array([lindex $element 0])"
	} else {
	    puts -nonewline $couplingFileId "0"
	}
#put new line
	puts $couplingFileId ""
    }
    close $couplingFileId
    puts "Finished collecting metrics (thanks for waiting)." 
    puts -nonewline "\nNOTE : Metrics can be found in a file $path/$projname-[string trim $access_option -]-coupling.csv."
    puts  " For best lisibility open it with Sun's StarCalc (Linux) or MS Excel (Windows) using comma separated columns."
}




proc extract_loc_na_nm {class_info access_option} {
    global db_classe
    global db_unions
    global db_attributes
    global db_methods

    set class_name [lindex $class_info 0] 
    set class_filename [lindex $class_info 3]

    set start_line [lindex $class_info 1]
    set start_line [lindex [extract_line_numbers [list $start_line {}]] 0]
    set end_line [lindex $class_info 2]
    set end_line [lindex [extract_line_numbers [list $end_line {}]] 0]

    set loc [expr $end_line - $start_line]
    

    switch -exact -- $access_option {
	-pub { set class_attributes [$db_attributes seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:4"]]; set class_methods [$db_methods seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:4"]] }
	-priv { set class_attributes [$db_attributes seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:1"]]; set class_methods [$db_methods seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:1"]] }
	-prot { set class_attributes [$db_attributes seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:2"]]; set class_methods [$db_methods seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line" "5 :0:2"]] }
	-all { set class_attributes [$db_attributes seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line"]]; set class_methods [$db_methods seq -col [list "0 =$class_name" 1 "3 =$class_filename" "4 -$start_line:$end_line"]]}
    }

    set na [llength $class_attributes]
    set nm [llength $class_methods]

    return "$loc,$na,$nm,"
}


proc find_DIT {} {
    global db_inheritance
    global DIT_array

    set list_base_classes {}
    set list_derived_classes {}

    if {![info exists db_inheritance]} {
	setLog "No inheritance in the project."
	return
    }

    set inheritance_info [$db_inheritance seq -col [list 0 1 2 3 4 5 8]]

    foreach entry $inheritance_info {
	set derived [lindex $entry 0]
	set base [lindex $entry 1]
	set inheritance($base) [lappend inheritance($base) $derived] 
	set list_base_classes [lappend list_base_classes $base]
	set list_derived_classes [lappend list_derived_classes $derived]
    }

    set list_base_classes [eliminate_duplicates_from_list $list_base_classes]
    set list_derived_classes [eliminate_duplicates_from_list $list_derived_classes]

    foreach base_class $list_base_classes {
	if {[lsearch $list_derived_classes $base_class] == -1 } {
	    set list_roots [lappend list_roots $base_class]
	}
    }


    foreach root $list_roots {
	set DIT_array($root) 0
	finish_finding_DIT inheritance $root 0
    }

}


proc finish_finding_DIT {inheritance_arr root_class dit_label} {
    global DIT_array
    upvar $inheritance_arr inheritance_array

    set dit_label [incr dit_label 1]

    if {[info exists inheritance_array($root_class)]} {
	foreach subclass $inheritance_array($root_class) {
	    if {![info exists DIT_array($subclass)]} {
		set DIT_array($subclass) $dit_label
	    } else {
		if {$dit_label > $DIT_array($subclass)} {
		    set DIT_array($subclass) $dit_label
		}
	    }
	    finish_finding_DIT inheritance_array $subclass $dit_label
	}
    }
}


# this procedure extracts class names from the file .cl and puts
# them in an array named vars_c


proc extract_classes {path projname} {
    global db_classe
    global vars_c
    global templates

    if {[file exists $path/${projname}.cl]} {
	set db_classe [dbopen nav_classe $path/${projname}.cl RDONLY 0644 btree]
    } else {
	setLog "File $path/${projname}.cl doesn't exist."
	return
    }

    foreach entry_c [$db_classe  seq] {
#puts -nonewline "."
	set subentry_c1 [lindex $entry_c 0]
	set subentry_c2 [lindex $entry_c 1]
	set class_c [lindex $subentry_c1 0]
	set template_args [lindex $subentry_c2 3]
# sets an array with class_c as name and 1 as value of an array element
	if {[string compare $template_args ""] == 0} {
	    set vars_c($class_c) 1
	} else {	    
	    set list_of_template_args [split [string trim $template_args <>] ,]
# have to extract class X, class Y parameters as a list of X,Y. discard the rest for the moment. in order to match template arguments names with template arguments types, we need all arguments, and not just class ones. array templates has it all. empty list in the end is for future template_args types.
    
	    set list_templ_args {}
	    set arg_counter 0
	    foreach element $list_of_template_args {
#array of template arg types. empty list will be constructed in check_for_classes
 
		set templates($class_c,$arg_counter) [list $element]
		incr arg_counter 1
		if {[lsearch $element class] != -1} {
		    #if you have something like this as template arguments : <class T,class SizeType=int,int IND_OFFSET=index_from_zero>, you don't want = and what's following it. So better get rid of it, you don't want it to be a part of the type.
		    regsub {([^=]*)(=)([^=]*)} $element {\1} element 
		    lappend list_templ_args [lindex $element 1]
		}
	    }
# if the value of an element in vars_c is 1, it's not a template. if it's 
# different than 1, it is in fact a list of template arguments of class type
	    set vars_c($class_c) $list_templ_args
	}
    }
}




proc is_template {my_class} {
    global vars_c
    if {[lsearch [array names vars_c] $my_class] != -1} {
	if {$vars_c($my_class) != 1} {
	    return 1
	}
    }
    return 0
}


# this procedure extracts typedefs names and types from the file .t
# and #puts them in vars_tn and vars_tt arrays, respectively. array
# names are set to 0, 1, 2 ..., to make it easier to extract type of 
# typedef when typedef's name is known. we could have used one array,
# and put typedef name as array name and typedef type as array value, 
# but there can be more than one occurence of a typedef's name, so
# we couldn't keep track of all of them. 


proc extract_typedefs {path projname} {
    global db_typedefs

    if {[file exists $path/${projname}.t]} {
	set db_typedefs [dbopen nav_typedefs $path/${projname}.t RDONLY 0644 btree]
    } else {
	setLog "File $path/${projname}.t doesn't exist"
	return
    }
}


proc extract_unions {path projname} {
    global db_unions
    if {[file exists $path/${projname}.un]} {
	set db_unions [dbopen nav_unions $path/${projname}.un RDONLY 0644 btree]
    } else {
	setLog "File $path/${projname}.un doesn't exist"
	return
    }
}




proc extract_included_files {projdir projname} {
#    global array_included_files
    if {[file exists $projdir/${projname}.iu]} {
	set db_included_files [dbopen nav_incl_files $projdir/${projname}.iu RDONLY 0644 btree] 
    } else {
	setLog "File $projdir/${projname}.iu doen't exist"
	return
    }

    set list_main_files {}
    foreach entry_if [$db_included_files seq] {
#puts -nonewline "."
	set sub_entry_if [lindex $entry_if 0]
	set included_file [lindex $sub_entry_if 0]
	set main_file [lindex $sub_entry_if 2]
	set array_incl_files($main_file) [lappend array_incl_files($main_file) $included_file]

#this is important, because we have one entry for each included file. here we collect all the files that include other files
	set list_main_files [lappend list_main_files $main_file] 
    }

#we could check every time when we append main_file to this list, if it's present in the list, or we can eliminate duplicates afterwards. we do it the latter way. 
    set list_main_files [eliminate_duplicates_from_list $list_main_files]

#for each main file call the procedure which does a search for all included files
    foreach main_file $list_main_files {
	finish_list_included_files $db_included_files $main_file $array_incl_files($main_file) 0
    }

}



#
# this procedure is called by extract_included_files for each main_file. 
# it starts with the initial list of included files found in the main_file.
# for each file, it finds its included files and appends them to
# the initial list. it proceeds in this manner until whole list is searched. 
# it's cheney's breadth-first algorithm. at the end, the 
# array_included_files(main_file) is set. 
#


proc finish_list_included_files {db_inc_files main_file list_included_files index} {
    global array_included_files
    
    if {$index <= [expr [llength $list_included_files] - 1]} {
	set inc_file [lindex $list_included_files $index]
	set base_inc_file [$db_inc_files seq -col [list 0 "2 =$inc_file"]]

#if we don't eliminate duplicates, the procedure is unable to finish.
	set base_inc_file [eliminate_duplicates_from_list $base_inc_file]
	foreach entry $base_inc_file {
#extract the included file
	    set included_file [lindex $entry 0]
#if the included file is already present in the list, don't consider it. if yes append it to the list of included files
	    if {[lsearch $list_included_files $included_file] == -1} { 
		set list_included_files [concat $list_included_files $included_file]
	    }
	}

#set index to the index of the next element in the list
	set index [expr $index + 1]
#recursive call with the updated index, in order to treat next included file for its included files
	finish_list_included_files $db_inc_files $main_file $list_included_files $index
    } else {
#if we reached the end of the list, we can set array_included_files with the complete list
	set array_included_files($main_file) [eliminate_duplicates_from_list $list_included_files]

    }  
}


proc extract_inheritance_info {path projname} {
    global db_inheritance
    global db_classe
    global parents_array

    if {[file exists $path/${projname}.in]} {
	set db_inheritance [dbopen inheritance $path/${projname}.in RDONLY 0644 btree]
    } else {
	setLog "File $path/${projname}.in doesn't exist"
	return
    }
        
    set all_classes [$db_classe seq -col [list 0 1 3 2]]

    foreach element $all_classes {
#puts -nonewline "."
	set my_class [lindex $element 0]
	set start_line [lindex $element 1]
	set start_line [lindex [extract_line_numbers [list $start_line {}]] 0]
	set end_line [lindex $element 2]
	set end_line [lindex [extract_line_numbers [list $end_line {}]] 0]
	set filename [lindex $element 3]

#it has to be this way because of check_type_for_typedefs
#we don't need parents_array element for a class that has no parents. So better check it before : 
	set parents_list [eliminate_duplicates_from_list [has_parents $my_class $start_line $end_line $filename {} 0]]
	if {$parents_list != {}} {
	    set parents_array($element) $parents_list
	}
    }
}


proc setLog {message} {
    global logFile
    global logFileError
    if {!$logFileError} {
	puts $logFile $message
    }

}




#  Main part of the program : input of the filename and directory. Call 
#  of the appropriate functions for extracting global lists.


if {[lindex $argv 1] == "-myhelp"} {
    puts "\t Usage: [lindex $argv 0] projectdir project_name acces_option\[-pub, -prot, -priv or -all\]. For help: [lindex $argv 0] -myhelp.  \n \t projectdir \t home directory of the Source Navigator project directory (.snprj)\n \t project_name \t name of the SN project (without .proj)\n \t -pub \t\t treat only public attributes or methods \n \t -priv \t\t treat only private attributes or methods \n \t -prot \t\t treat only protected attributes or methods \n \t -all \t\t treat all attributes or methods \n \t -myhelp \t this help message"
    exit
}

if {$argc != 4} {
#    puts "Usage: $argv0 projectdir project access_option\[-pub, -prot, -priv or -all\]. For help: $argv0 -myhelp"
	puts "Usage: [lindex $argv 0] projectdir project_name access_option\[-pub, -prot, -priv or -all\]. For help: [lindex $argv 0] -myhelp"
    exit
}


#set tcl_file $argv0


#set projdir [lindex $argv 0]/.snprj
#set projname [lindex $argv 1]
#set access_option [string tolower [lindex $argv 2]]

set projdir [lindex $argv 1]/.snprj
set projname [lindex $argv 2]
set access_option [string tolower [lindex $argv 3]]

if { [regexp {^(-pub|-prot|-priv|-all)$} $access_option] == 0 } {
    puts "Wrong option : $access_option"; puts "Usage: [lindex $argv 0] projectdir project_name acces_option\[-pub, -prot, -priv or -all\]"
    exit
}

if {![file exists [lindex $argv 1]/${projname}.proj]} {
    puts "Wrong filename. Make sure that the project directory and the project filename are right."
    puts "Usage: [lindex $argv 0] projectdir project_name acces_option\[-pub, -prot, -priv or -all\]"
    exit
}



global logFileError
global logFile

global db_attributes
global db_methods


set logFileError [catch {set logFile [open $projdir/$projname\_coupling.log w 0600]}]


puts "Extracting classes ... (please wait)"
extract_classes $projdir $projname
puts "Extracting typedefs ... (please wait)"
extract_typedefs $projdir $projname
puts "Extracting included files ... (please wait)"
extract_included_files $projdir $projname
puts "Extracting inheritance info ... (please wait)"
extract_inheritance_info $projdir $projname
puts "Extracting unions ... (please wait)"
extract_unions $projdir $projname


set attributes {}
set methods {}

if {[file exists $projdir/${projname}.iv]} {
    set db_attributes [dbopen nav_attributes $projdir/${projname}.iv RDONLY 0644 btree]    
    set attributes [$db_attributes seq -list]
} else {
    setLog "File $projdir/${projname}.iv doesn't exist"
}


if {[file exists $projdir/${projname}.md]} {
   set db_methods [dbopen nav_methods $projdir/${projname}.md RDONLY 0644 btree]
    set methods [$db_methods seq -list]
} else {
    setLog "File $projdir/${projname}.md doesn't exist"
}



switch -exact -- $access_option {
    -pub { extract_attribute_type [select_public $db_attributes]; extract_method_return_and_argument_types [select_public $db_methods]}
    -priv { extract_attribute_type [select_private $db_attributes]; extract_method_return_and_argument_types [select_private $db_methods] }
    -prot { extract_attribute_type [select_protected $db_attributes]; extract_method_return_and_argument_types [select_protected $db_methods] }
    -all {extract_attribute_type $attributes; extract_method_return_and_argument_types $methods}
    default {puts "Wrong option : $access_option"; puts "Usage: [lindex $argv 0] projectdir project_name acces_option\[-pub, -prot, -priv or -all\]"; exit}
}


finish_template_arg_coupling
find_DIT
quantify_coupling $projdir $projname $access_option

exit










